import spacy
from spacy.pipeline import Sentencizer


class CustomSentencizer():

  @staticmethod
  def __custom_seg(doc):
    length = len(doc)
    for index, token in enumerate(doc):
        if (token.text == ';' and index!=(length - 1)):
            doc[index+1].sent_start = True
    return doc

  def __init__(self):
    self.nlp = spacy.load("fr_core_news_sm")
    self.nlp.add_pipe(self.__custom_seg, before='parser')

  
  def sentencize(self, text):
    try :
        doc = self.nlp(text)
        return [str(s) for s in doc.sents]
    except Exception as e:
        return []



from os import rename
from pandas.io.parsers import ParserBase
from ..enrichment.sentencizer import CustomSentencizer
from ..kairntech import Annotator, KairntechClient
import pandas as pd
import numpy as np
import uuid
import json
from typing import Any, Callable, Dict, List, Union
from .meta_column import MetaColumn
from ..processors import processors
from ..pipeline.task import Task
from pathlib import Path
from datetime import datetime


class FileFormater(Task):

    def __init__(self,
                name,
                 verbatim_column:str,
                 date_column:str,
                 id_column:str,
                 meta_columns:List[MetaColumn],
                 drop_empty_verbatim:bool=True,
                 encoding:str='utf-8',
                 generate_id=False, 
                 csv_separator=","):
        super().__init__(name)
        self.verbatim_column = verbatim_column
        self.date_column = date_column
        self.id_column = id_column
        self.meta_columns = meta_columns
        self.drop_empty_verbatim = drop_empty_verbatim
        self.encoding = encoding
        self.generate_id = generate_id
        self.csv_separator = csv_separator
        self.output_extension = ".json"

    def __rename_columns(self, original_columns):
        columns = []
        for c in original_columns:
            if c==self.id_column:
                columns.append("id")
            if c==self.date_column:
                columns.append("dateInterview")
            if c==self.verbatim_column:
                columns.append("text")
            elif c in [m.column_name for m in self.meta_columns]:
                meta_column = list(filter(lambda x: x.column_name==c,self.meta_columns))[0]
                if meta_column.rename_column:
                    columns.append(meta_column.new_column_name)
                else:
                    columns.append(meta_column.column_name)
        return columns
        
    def __generate_id(self):
        return str(uuid.uuid4())

    @property
    def __columns_to_keep(self):
        return [self.id_column, self.date_column, self.verbatim_column] + [m.column_name for m in self.meta_columns]


    def __get_dataframe_from_path(self,filepath:Path)->pd.DataFrame:
        if filepath.suffix == ".xlsx":
            data = pd.read_excel(filepath, parse_dates=[self.date_column], encoding=self.encoding)
        elif filepath.suffix== ".csv":
            data = pd.read_csv(filepath, parse_dates=[self.date_column], encoding=self.encoding, sep=self.csv_separator)
        else :
            raise Exception("only csv and xlsx format are supported. Extension is " + filepath.suffix)
        if self.generate_id:
            data[self.id_column] = data[data.columns[0]].apply(lambda x: self.__generate_id())
        return data

    def __list_of_filters(self):
        meta_columns = [m.new_column_name if m.rename_column else m.column_name for m in self.meta_columns]
        return meta_columns

    def __format_pcx(self, data):
        formatted_data = []
        filters = self.__list_of_filters()
        for row in data:
            formatted_row = {
                "id": row["id"],
                "dateInterview": row["dateInterview"].strftime("%Y-%m-%d %H:%M:%S"),
                "text": row["text"],
                "data": {f:row[f] for f in filters}
            }
            formatted_data.append(formatted_row)
        return formatted_data
            

    def run(self, *input)->str:
        """
        Format the given file to json format
        """
        data = self.__get_dataframe_from_path(input[0])
        # keep only the good columns
        data = data[self.__columns_to_keep]
        # drop empty verbatim if needed
        if self.drop_empty_verbatim:
            data[self.verbatim_column].replace('', np.nan,inplace=True)
            data.dropna(subset=[self.verbatim_column],inplace=True)
        # fill na if needed
        for meta in self.meta_columns:
            if meta.fill_na:
                data[meta.column_name].fillna(meta.na_filler)
        # rename columns if needed
        data.columns = self.__rename_columns(data.columns)
        pcx_formatted_data = self.__format_pcx(data.to_dict(orient="records"))
        return json.dumps(pcx_formatted_data)



from ..pipeline.task import Task
from ..kairntech import KairntechClient, Annotator
from ..enrichment.classifier import annnotate_sentiment_and_topic
import json


class KairntechTask(Task):

    def __init__(self, name, topic_annotator: Annotator, sentiment_annotator: Annotator, sentencizer=None, sentiment_meta=True, topic_meta=True, verbatim_column="text"):
        super().__init__(name)
        self.output_extension = ".json"
        self.topic_annotator = topic_annotator
        self.sentiment_annotator = sentiment_annotator
        self.sentencizer = sentencizer
        self.sentiment_meta = sentiment_meta
        self.topic_meta = topic_meta
        self.verbatim_column = verbatim_column

    def run(self, *input):
        input_file = input[0]
        with open(input_file, "r") as f:
            data = json.load(f)
        for row in data:
            classifier_result = annnotate_sentiment_and_topic(row[self.verbatim_column],
                                                              self.topic_annotator,
                                                              self.sentiment_annotator, self.sentencizer)
            if classifier_result[1] != None:
                row["tonality"] = classifier_result[1].lower()
                if self.sentiment_meta:
                    row["data"]["Sentiment"] = classifier_result[1].lower()
            if len(classifier_result[0]):
                row["themes"] = classifier_result[0]
                if self.topic_meta:
                    row["data"]["Topic"] = [topic["key"]
                                            for topic in classifier_result[0]]
        return json.dumps(data)



from ..pre_processing import MetaColumn
import pandas as pd
from typing import List


def bootstrap_meta_from_file(filepath, mandatory_columns: List, encoding="utf-8", csv_separator=","):
    if filepath.split(".")[1] == "xlsx":
        data = pd.read_excel(filepath, encoding=encoding, n_rows=2)
    elif filepath.split(".")[1] == "csv":
        data = pd.read_csv(filepath, encoding=encoding, sep=csv_separator, nrows=2)
    else:
        raise Exception("only csv and xlsx format are supported")
    data.drop(mandatory_columns, axis=1, inplace=True)
    return [MetaColumn(c).serialize() for c in data.columns]


def list_column_from_file(filepath, encoding="utf-8", csv_separator=","):
    if filepath.split(".")[1] == "xlsx":
        data = pd.read_excel(filepath, encoding=encoding, n_rows=2)
    elif filepath.split(".")[1] == "csv":
        data = pd.read_csv(filepath, encoding=encoding, sep=csv_separator, nrows=2)
    else:
        raise Exception("only csv and xlsx format are supported")
    return data.columns.tolist()



from typing import Any, Callable, Dict, List, Union
import json

class MetaColumn():
    
    def __init__(self,
                column_name:str,
                rename_column:Union[bool,str]=False,
                fill_na:Union[bool,Any]=False):
        """
        Initialize a new meta column

        Args:
        column_name : name of the target column in the dataset
        rename_column: False by default, to rename column provide the new column name
        fill_na : if not False na values will be replaced by the value provided
        """
        self.column_name = column_name
        if rename_column:
            self.rename_column = True
            self.new_column_name = rename_column
        else:
            self.rename_column = False
        if fill_na:
            self.fill_na = True
            self.na_filler = fill_na
        else:
            self.fill_na = False



import requests
import json

class KairntechClient():

    @staticmethod
    def get_headers(token=None, content_type="application/json"):
        headers = {"accept": "application/json",
                   "Content-Type": content_type}
        if token:
            headers["Authorization"] = "Bearer " + token
        return headers

    def __get_headers(self, token=None, content_type="application/json"):
        headers = {"accept": "application/json",
                   "Content-Type": content_type}
        if token:
            headers["Authorization"] = "Bearer " + token
        return headers

    def __login(self, email, password):
        headers = self.__get_headers()
        payload = json.dumps({"email": email, "password": password})
        r = requests.post(f"{self.api_url}auth/login",
                          headers=headers, data=payload)
        return r.json()["access_token"]

    def __init__(self, email, password, api_url=f"https://sherpa.kairntech.com/api/"):
        self.api_url = api_url
        self.token = self.__login(email, password)

    def get_projects(self):
        headers = self.__get_headers(self.token)
        r = requests.get(
            f"{self.api_url}projects?computeMetrics=false", headers=headers)
        results = r.json()
        return [x["name"] for x in results]

    def get_annotators(self, project_name):
        headers = self.get_headers(self.token)
        r = requests.get(
            f"{self.api_url}projects/{project_name}/annotators_by_type", headers=headers)
        results = r.json()
        if "learner" in results:
            return [(x["label"], x["name"])
                    for x in results["learner"]]
        else:
            return []






from .client import KairntechClient
import requests
from time import sleep


class Annotator():

    def __init__(self, client: KairntechClient, project_name: str, annotator_name: str, sleep_between=0):
        self.project_name = project_name
        self.annotator_name = annotator_name
        self.client = client
        self.sleep_between = sleep_between

    def annotate(self, document: str):
        document = document.encode("utf-8")
        headers = KairntechClient.get_headers(self.client.token, "text/plain")
        r = requests.post(
            f"{self.client.api_url}projects/{self.project_name}/annotators/{self.annotator_name}/_annotate", headers=headers, data=document)
        results = []
        for c in r.json()["categories"]:
            
            if "label" in c:
                results.append(c["label"])
            else:
                results.append(c["labelName"])
        sleep(self.sleep_between)
        return results



from .task import Task, FileReader
from typing import List
from pathlib import Path
import logging


class TaskRunner():

    def __init__(self, name: str, **kwargs):
        self.name = name
        self.args = kwargs

    def output_filename(self, task):
        return self.name + "-" + task.name + "-" + "-".join([str(arg) for arg in self.args.values()]) + task.output_extension

    def is_task_completed(self, task):
        raise NotImplementedError

    def run_task(self, task: Task, input_task: List[Task] = []):
        raise NotImplementedError

    def check_input_tasks(input_tasks: List[Task]):
        raise NotImplementedError


class FileTaskRunner(TaskRunner):

    def __init__(self, name: str, folder: str, verbose=False, **kwargs):
        super().__init__(name, **kwargs)
        self.folder = folder
        self.verbose = verbose

    def is_task_completed(self, task: Task):
        return self.output_filename(task).exists()

    def output_filename(self, task):
        return Path(self.folder, super().output_filename(task))

    def check_input_tasks(self, input_tasks: List[Task]):
        for task in input_tasks:
            if not self.is_task_completed(task):
                return False
        return True

    def run_task(self, task: Task, input_tasks: List[Task] = []):
        if self.verbose:
            logging.getLogger().setLevel(logging.DEBUG)
        logging.info(f"Start running {task.name}")
        if self.is_task_completed(task):
            logging.info(f"Task {task.name} is already completed")
            return 0
        if self.check_input_tasks(input_tasks):
            task_output = task.run(*[self.output_filename(task)
                                     for task in input_tasks])
            if isinstance(task, FileReader):
                with open(self.output_filename(task), "wb") as f:
                    f.write(task_output)
            else:
                with open(self.output_filename(task), "w") as f:
                    f.write(task_output)
        else:
            logging.info(
                f"Task dependencies for {task.name} are not completed")
            return 0
        logging.info(f"Task {task.name} completed")



from .pcx import PushPCXTask
from .pipeline.pipeline import Pipeline
from .pipeline.task import Task, Reader, FileReader
from .pipeline.task_runner import TaskRunner, FileTaskRunner
from .kairntech import KairntechClient, Annotator
from .enrichment.kairntech_task import KairntechTask
from .pre_processing import FileFormater, MetaColumn
from .processors.recoder import Recoder




from .task import Reader, Task
from .task_runner import TaskRunner
from typing import List

class Pipeline():

  def __init__(self, task_runner: TaskRunner, tasks: List[Task]):
    self.task_runner = task_runner
    self.tasks = tasks

  def run(self):
    for i, task in enumerate(self.tasks):
      if isinstance(task,Reader):
        self.task_runner.run_task(self.tasks[i])
      else:
        self.task_runner.run_task(self.tasks[i], [self.tasks[i-1]])



def recod_nps(value):
  value = int(value)
  if value>=0 and value<=6:
    return "Détracteur"
  elif value>=7 and value<=8:
    return "Neutre"
  elif value>=9:
    return "Promoteur"

def lowercase(value):
  value = str(value)
  return value.lower()

def uppercase(value):
  value = str(value)
  return value.upper()

def capitalize(value):
  value = str(value)
  return value.capitalize()


from ..kairntech import KairntechClient, Annotator
from .sentencizer import CustomSentencizer

def consolidate_sentiment(results:list):
    consolidated = {f"{x['key']}||{x['tonality']}" for x in results}
    consolidated = [{"key": x[0], "tonality": x[1]}
                    for x in [x.split("||") for x in consolidated]]
    global_tonality=None
    for topic in consolidated:
        if global_tonality==None:
            global_tonality = topic["tonality"]
            continue
        if global_tonality!=topic["tonality"]:
            global_tonality="mixed"
            break
    return consolidated, global_tonality


def annnotate_sentiment_and_topic(doc: str,
                                  topic_annotator : Annotator,
                                  sentiment_annotator: Annotator,
                                  sentencizer=None):
    results = []
    doc = str(doc)
    if sentencizer:
        sentences = sentencizer.sentencize(doc)
        for sentence in sentences:
            topics = topic_annotator.annotate(sentence)
            if len(topics):
                sentiment = sentiment_annotator.annotate(sentence)
                if not len(sentiment):
                    sentiment = None
                else:
                    sentiment = sentiment[0].lower()
                    for code in [{"key": t, "tonality": sentiment} for t in topics]:
                        results.append(code)
    else:
        topics = topic_annotator.annotate( doc)
        sentiment = sentiment_annotator.annotate(doc)
        if not len(sentiment):
            sentiment = None
        else:
            sentiment = sentiment[0].lower()
            for code in [{"key": t, "tonality": sentiment} for t in topics]:
                results.append(code)
    return consolidate_sentiment(results)



from typing import List
from pathlib import Path


class Task():

  def __init__(self, name):
    self.name = name
    self.output_extension = ".txt"

  def run(self, *input):
    raise NotImplementedError

class DummyTask(Task):

  def __init__(self, name):
    super().__init__(name)

  def run(self, *input):
    return self.name


class Reader(Task):

  def __init__(self, name):
    super().__init__(name)

  def run(self):
    raise NotImplementedError



class FileReader(Reader):

  def __init__(self, name, filename):
    super().__init__(name)
    self.filename = filename
    self.output_extension = Path(filename).suffix

  def run(self, *_input):
    with open(self.filename, "rb") as f:
      content = f.read()
    return content
   


from os import stat_result
from verbatim_processor.helpers.chunk import chunk
from verbatim_processor.pipeline.task import Task
import json
import requests
import logging
from verbatim_processor.helpers import chunk
import math


class PushPCXTask(Task):

    def __init__(self, name, token, survey_name, chunk_size=1000, pcx_api_url="https://collector.bva-cx-insights.com/api/1.0/verbatim/"):
        super().__init__(name)
        self.output_extension = ".json"
        self.pcx_api_url = pcx_api_url
        self.token = token
        self.survey_name = survey_name
        self.chunk_size = chunk_size

    def push_data(self, data):
        headers = {"Authorization": "Token " + self.token}
        r = requests.post(self.pcx_api_url, json=data, headers=headers)
        logging.info(
            f"Push to pcx ... Status code: {r.status_code} Text: {r.text}")
        if r.status_code != 200:
            logging.error(
                f"Push to PCX failed... Status code: {r.status_code} Text:{r.text}")
        return r.status_code, r.text

    def run(self, *input):
        with open(input[0]) as f:
            data = json.load(f)
        for row in data:
            row["surveyName"] = self.survey_name
        # the dataset is splits in chunks to not overstress the api
        # responses array will store the api responses for each chunk
        responses = []
        for subset in chunk(data, self.chunk_size):
            status_code, text = self.push_data(subset)
            responses.append(
                {"chunk": subset, "statusCode": status_code, "text": text})
        return json.dumps(responses)



import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent
print(find_packages(exclude=("verbatim_processor.test.*","verbatim_processor.test")))
# The text of the README file
README = (HERE / "README.md").read_text()
# This call to setup() does all the work
setup(
    name="pcx-utilities",
    version="0.1.0",
    description="A package to easily create pipeline for bva-cx-insights.com",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://bva-cx-insights.com",
    author="Adrien Liard",
    author_email="adrien.liard@bva-group.com",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
    ],
   packages=find_packages(exclude=("verbatim_processor.test.*","verbatim_processor.test")),
    include_package_data=True,
    install_requires=["pandas", "requests", "spacy"]
)


from typing import Any, Callable, Dict, List, Union
import json
import logging
from verbatim_processor.pipeline.task import Task


class Recoder(Task):

    def __init__(self,
                name:str,
                 column_name: str,
                 processor: Union[Callable, Dict],
                 in_place: bool=True,
                 new_column_name: str=None,
                 drop_column:bool=False):
        """
        Recoders are useful to modify values in a column

        Args:
        column_name: name of the target column
        processor: A callable from the processors module or a Dict ({"old value": "new value"})
        in_place: if true the column will be replaced, new_column_name should then not be None
        new_column_name: new column will be created with this name, used only if in_place==False
        drop_column : if True the old column is dropped, used only if in_place==Fals
       
        """
        super().__init__(name)
        self.output_extension = ".json"
        self.column_name = column_name
        self.processor = processor
        self.in_place = in_place
        self.new_column_name = new_column_name
        self.drop_column = drop_column

        if in_place==False and new_column_name==None:
            raise Exception("A recoder can't have in_place==True without new_column_name")

        if type(self.processor) == dict:
            self.processor_type = "dict"
        else:
            self.processor_type = "func"
        
    def apply(self, value):
            if self.processor_type == "dict":
                if value in self.processor:
                    return self.processor[value]
                else:
                    logging.warning(f"Error in recoder {self.column_name} Value '{value}' not in dict")
                    return None
            try:
                return self.processor(value)
            except Exception as error:
                logging.warning(f"Error in recoder {self.column_name}")
                if self.raise_errors:     
                    raise error
                else:
                    return None 

    def run(self, *input):
        with open(input[0]) as f:
            data = json.load(f)
        for row in data:
            if self.in_place:
                row["data"][self.column_name] = self.apply(row["data"][self.column_name])
            elif self.drop_column==False:
                row["data"][self.new_column_name] = self.apply(row["data"][self.column_name])
            elif self.drop_column:
                value = row["data"].pop(self.column_name, None)
                if value!=None:
                    row["data"][self.new_column_name] = self.apply(value)
        return json.dumps(data)

